import json
import os
import zipfile
from pathlib import Path
from typing import Tuple

from caseconverter import snakecase
from google.api_core.exceptions import PermissionDenied
from google.cloud import scheduler_v1
from google.cloud.exceptions import NotFound
from google.cloud.functions_v1 import CloudFunctionsServiceClient

from wanna.core.deployment.io import IOMixin
from wanna.core.deployment.models import (
    AlertPolicyResource,
    CloudFunctionResource,
    CloudSchedulerResource,
    LogMetricResource,
)
from wanna.core.deployment.monitoring import MonitoringMixin
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.utils import templates
from wanna.core.utils.gcp import is_gcs_path

logger = get_logger(__name__)


class VertexSchedulingMixIn(MonitoringMixin, IOMixin):
    def upsert_cloud_scheduler(
        self, function: Tuple[str, str], resource: CloudSchedulerResource, version: str, env: str
    ) -> None:
        client = scheduler_v1.CloudSchedulerClient(credentials=self.credentials)
        parent = f"projects/{resource.project}/locations/{resource.location}"
        job_id = f"{resource.name}-{env}"
        job_name = f"{parent}/jobs/{job_id}"
        function_name, function_url = function

        logger.user_info(f"Deploying {resource.name} cloud scheduler with version {version} to env {env}")

        http_target = {
            "uri": function_url,
            "body": json.dumps(resource.body, separators=(",", ":")).encode(),
            "headers": {
                "Content-Type": "application/octet-stream",
                "User-Agent": "Google-Cloud-Scheduler",
                "Wanna-Pipeline-Version": version,
            },
            "oidc_token": {
                "service_account_email": resource.cloud_scheduler.service_account or resource.service_account,
                # required scope https://developers.google.com/identity/protocols/oauth2/scopes#cloudfunctions
                # "scope": "https://www.googleapis.com/auth/cloud-platform"
            },
        }

        job = {
            "name": job_name,
            "description": f"wanna {resource.name} scheduler for  {env} pipeline",
            "http_target": http_target,
            "schedule": resource.cloud_scheduler.cron,
            "time_zone": resource.cloud_scheduler.timezone,
            # TODO: "retry_config" ,
            # TODO: "attempt_deadline"
        }

        try:
            _ = client.get_job({"name": job_name})
            logger.user_info(f"Found {job_name} cloud scheduler job, updating it")
            client.update_job({"job": job, "update_mask": {"paths": ["schedule", "http_target", "time_zone"]}})

        except NotFound:
            # Does not exist let's create it
            logger.user_info(f"Creating {job_name} with deployment manifest for {env} with version {version}")
            client.create_job({"parent": parent, "job": job})

        logging_metric_ref = f"{job_id}-cloud-scheduler-errors"
        gcp_resource_type = "cloud_scheduler_job"
        self.upsert_log_metric(
            LogMetricResource(
                project=resource.project,
                location=resource.location,
                name=logging_metric_ref,
                filter_=f"""
            resource.type="{gcp_resource_type}" AND severity >= WARNING AND resource.labels.job_id="{job_id}"
            """,
                description=f"Log metric for {resource.name} cloud scheduler job",
            )
        )

        self.upsert_alert_policy(
            AlertPolicyResource(
                logging_metric_type=logging_metric_ref,
                resource_type=gcp_resource_type,
                project=resource.project,
                location=resource.location,
                name=f"{job_id}-cloud-scheduler-alert-policy",
                display_name=f"{job_id}-cloud-scheduler-alert-policy",
                labels=resource.labels,
                notification_channels=resource.notification_channels,
            )
        )

    def upsert_cloud_function(self, resource: CloudFunctionResource, version: str, env: str) -> Tuple[str, str]:
        logger.user_info(f"Deploying {resource.name} cloud function with version {version} to env {env}")
        parent = f"projects/{resource.project}/locations/{resource.location}"
        pipeline_functions_dir = resource.build_dir / "functions"
        os.makedirs(pipeline_functions_dir, exist_ok=True)
        local_functions_package = pipeline_functions_dir / "package.zip"
        functions_gcs_path_dir = f"{resource.resource_root}/functions"
        functions_gcs_path = f"{functions_gcs_path_dir}/package.zip"
        function_name = f"{resource.name}-{env}"
        function_path = f"{parent}/functions/{function_name}"

        cloud_function = templates.render_template(
            Path("scheduler_cloud_function.py"),
            manifest=resource.template_vars,
            labels=json.dumps(resource.labels, separators=(",", ":")),
        )

        requirements = templates.render_template(
            Path("scheduler_cloud_function_requirements.txt"), manifest=resource.template_vars
        )

        with zipfile.ZipFile(local_functions_package, "w") as z:
            z.writestr("main.py", cloud_function)
            z.writestr("requirements.txt", requirements)

        if not is_gcs_path(functions_gcs_path_dir):
            os.makedirs(functions_gcs_path_dir, exist_ok=True)

        self.upload_file(str(local_functions_package), functions_gcs_path)

        cf = CloudFunctionsServiceClient(credentials=self.credentials)
        function_url = f"https://{resource.location}-{resource.project}.cloudfunctions.net/{function_name}"
        function = {
            "name": function_path,
            "description": f"wanna {resource.name} function for {env} pipeline",
            "source_archive_url": functions_gcs_path,
            "entry_point": "process_request",
            "runtime": "python39",
            "https_trigger": {
                "url": function_url,
            },
            "service_account_email": resource.service_account,
            "labels": resource.labels,
            "environment_variables": {snakecase(k).upper(): v for k, v in resource.env_params.items()}
            # TODO: timeout
        }

        try:
            cf.get_function({"name": function_path})
            cf.update_function({"function": function}).result()
        # it can raise denied on resource 'projects/{project_id}/locations/{loaction}/functions/{function_name}'
        # (or resource may not exist).
        except (NotFound, PermissionDenied):
            cf.create_function({"location": parent, "function": function}).result()

        logging_metric_ref = f"{function_name}-cloud-function-errors"
        gcp_resource_type = "cloud_function"
        self.upsert_log_metric(
            LogMetricResource(
                name=logging_metric_ref,
                project=resource.project,
                location=resource.location,
                filter_=f"""
                resource.type="{gcp_resource_type}" AND severity >= WARNING
                AND resource.labels.function_name="{function_name}"
                """,
                description=f"Log metric for {function_name} cloud function executions",
            )
        )

        self.upsert_alert_policy(
            AlertPolicyResource(
                name=f"{function_name}-cloud-function-alert-policy",
                project=resource.project,
                location=resource.location,
                logging_metric_type=logging_metric_ref,
                resource_type=gcp_resource_type,
                display_name=f"{function_name}-cloud-function-alert-policy",
                labels=resource.labels,
                notification_channels=resource.notification_channels,
            )
        )

        return (
            function_path,
            function_url,
        )



import typer

from wanna.cli.plugins.components_plugin import ComponentsPlugin
from wanna.cli.plugins.job_plugin import JobPlugin
from wanna.cli.plugins.managed_notebook_plugin import ManagedNotebookPlugin
from wanna.cli.plugins.notebook_plugin import NotebookPlugin
from wanna.cli.plugins.pipeline_plugin import PipelinePlugin
from wanna.cli.plugins.tensorboard_plugin import TensorboardPlugin


class PluginRunner:
    def __init__(self) -> None:
        self.app = typer.Typer(rich_markup_mode="rich", help="Complete MLOps framework for Vertex-AI")

        typers = [
            ("pipeline", PipelinePlugin()),
            ("job", JobPlugin()),
            ("notebook", NotebookPlugin()),
            ("tensorboard", TensorboardPlugin()),
            ("managed-notebook", ManagedNotebookPlugin()),
            ("components", ComponentsPlugin()),
        ]
        for name, subcommand in typers:
            self.app.add_typer(subcommand.app, name=name, help=subcommand.__doc__)



import json
from datetime import datetime as dt

from google.cloud import aiplatform_v1, storage


def main(event, context):
    path = event["id"].rsplit("/", 1)[0].split("/", 1)[1]
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(event["bucket"])
    blob = bucket.blob(path)
    data = blob.download_as_text().splitlines()

    client = aiplatform_v1.PipelineServiceClient(
        client_options={"api_endpoint": "europe-west1-aiplatform.googleapis.com"},
    )

    for line in data:
        entry = json.loads(line)
        timestamp = entry["jsonPayload"]["startTime"]
        project_id = entry["logName"].split("/")[1]
        location = entry["resource"]["labels"]["location"]
        pipeline_id = entry["resource"]["labels"]["pipeline_job_id"]
        name = f"projects/{project_id}/locations/{location}/pipelineJobs/{pipeline_id}"
        delta = dt.now() - dt.strptime(timestamp[:-8].replace("T", " "), "%Y-%m-%d %H:%M:%S")

        request = aiplatform_v1.GetPipelineJobRequest(name=name)
        pipeline = client.get_pipeline_job(request=request)
        if delta.total_seconds() > 3600 * float(
            pipeline.labels["wanna_sla_hours"].replace("_", ".")
        ):  # 60 seconds * 60 minutes
            request = aiplatform_v1.CancelPipelineJobRequest(name=name)
            client.cancel_pipeline_job(request=request)



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, EmailStr, Extra, validator

from wanna.core.utils import validators


class BaseInstanceModel(BaseModel, extra=Extra.ignore, validate_assignment=True):
    name: str
    project_id: str
    zone: Optional[str]
    region: Optional[str]
    labels: Optional[Dict[str, str]]
    description: Optional[str]
    service_account: Optional[EmailStr]
    network: Optional[str]
    bucket: Optional[str]
    tags: Optional[List[str]]
    metadata: Optional[Dict[str, Any]]

    _project_id = validator("project_id", allow_reuse=True)(validators.validate_project_id)
    _zone = validator("zone", allow_reuse=True)(validators.validate_zone)
    _region = validator("region", allow_reuse=True)(validators.validate_region)
    _network = validator("network", allow_reuse=True)(validators.validate_network_name)
    _labels = validator("labels", allow_reuse=True)(validators.validate_labels)



import os
import shutil
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from caseconverter import kebabcase
from dirhash import dirhash
from google.api_core.client_options import ClientOptions
from google.api_core.future.polling import DEFAULT_POLLING
from google.api_core.operation import Operation
from google.cloud.devtools import cloudbuild_v1
from google.cloud.devtools.cloudbuild_v1.services.cloud_build import CloudBuildClient
from google.cloud.devtools.cloudbuild_v1.types import Build, BuildOptions, BuildStep, Source, StorageSource
from google.protobuf.duration_pb2 import Duration  # pylint: disable=no-name-in-module
from python_on_whales import Image, docker

from wanna.core.deployment.models import PushMode
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.docker import (
    DockerBuildConfigModel,
    DockerImageModel,
    DockerModel,
    ImageBuildType,
    LocalBuildImageModel,
    NotebookReadyImageModel,
    ProvidedImageModel,
)
from wanna.core.models.gcp_profile import GCPProfileModel
from wanna.core.utils import loaders
from wanna.core.utils.credentials import get_credentials
from wanna.core.utils.env import cloud_build_access_allowed, gcp_access_allowed
from wanna.core.utils.gcp import convert_project_id_to_project_number, make_tarfile, upload_file_to_gcs
from wanna.core.utils.templates import render_template

logger = get_logger(__name__)


class DockerClientException(Exception):
    pass


class DockerService:
    def __init__(
        self,
        docker_model: DockerModel,
        gcp_profile: GCPProfileModel,
        version: str,
        work_dir: Path,
        wanna_project_name: str,
        quick_mode: bool = False,  # just returns tags but does not build
    ):
        self.image_models = docker_model.images
        self.image_store: Dict[str, Tuple[DockerImageModel, Optional[Image], str]] = {}

        # Artifactory mirrors to different registry/projectid/repository combo
        registry_suffix = os.getenv("WANNA_DOCKER_REGISTRY_SUFFIX")
        self.docker_registry_suffix = f"{registry_suffix}/" if registry_suffix else ""
        self.docker_registry = (
            os.getenv("WANNA_DOCKER_REGISTRY")
            or docker_model.registry
            or gcp_profile.docker_registry
            or f"{gcp_profile.region}-docker.pkg.dev"
        )
        docker_repository = os.getenv("WANNA_DOCKER_REGISTRY_REPOSITORY") or docker_model.repository
        self.docker_repository = docker_repository if docker_repository else gcp_profile.docker_repository
        self.docker_project_id = os.getenv("WANNA_DOCKER_REGISTRY_PROJECT_ID") or gcp_profile.project_id

        self.version = version
        self.work_dir = work_dir
        self.build_dir = self.work_dir / "build" / "docker"
        self.wanna_project_name = wanna_project_name
        self.project_id = gcp_profile.project_id
        self.location = gcp_profile.region
        self.docker_build_config_path = os.getenv("WANNA_DOCKER_BUILD_CONFIG", self.work_dir / "dockerbuild.yaml")
        self.build_config = self._read_build_config(self.docker_build_config_path)
        self.cloud_build_timeout = docker_model.cloud_build_timeout
        self.cloud_build = (
            False if not gcp_access_allowed or not cloud_build_access_allowed else docker_model.cloud_build
        )

        self.cloud_build_workerpool = docker_model.cloud_build_workerpool
        self.cloud_build_workerpool_location = docker_model.cloud_build_workerpool_location or self.location
        self.bucket = gcp_profile.bucket
        self.quick_mode = quick_mode
        assert self.cloud_build or self._is_docker_client_active(), DockerClientException(
            "You need running docker client on your machine to use WANNA cli with local docker build"
        )

    def _read_build_config(self, config_path: Union[Path, str]) -> Union[DockerBuildConfigModel, None]:
        """
        Reads the DockerBuildConfig from local file.
        If the file does not exist, return None.

        Args:
            config_path:

        Returns:
            DockerBuildConfigMode
        """
        if os.path.isfile(config_path):
            with open(config_path) as file:
                # Load workflow file
                build_config_dict = loaders.load_yaml(file, self.work_dir)
            build_config = DockerBuildConfigModel.parse_obj(build_config_dict)
            return build_config
        return None

    @staticmethod
    def _is_docker_client_active() -> bool:
        """
        Method to find out if you have a running docker client on your machine.

        Returns:
            True if docker client found, False otherwise
        """
        return docker.info().id is not None

    def _build_image(
        self, context_dir, file_path: Path, tags: List[str], docker_image_ref: str, **build_args
    ) -> Union[Image, None]:

        should_build = self._should_build_by_context_dir_checksum(self.build_dir / docker_image_ref, context_dir)

        if should_build and not self.quick_mode:
            if self.cloud_build:
                logger.user_info(text=f"Building {docker_image_ref} docker image in GCP Cloud build")
                op = self._build_image_on_gcp_cloud_build(
                    context_dir=context_dir, file_path=file_path, docker_image_ref=docker_image_ref, tags=tags
                )
                project = convert_project_id_to_project_number(self.project_id)
                build_id = op.metadata.build.id
                base = "https://console.cloud.google.com/cloud-build/builds"
                if self.cloud_build_workerpool:
                    link = base + f";region={self.cloud_build_workerpool_location}/{build_id}?project={project}"
                else:
                    link = base + f"/{build_id}?project={project}"
                try:
                    op.result()
                    self._write_context_dir_checksum(self.build_dir / docker_image_ref, context_dir)
                except:
                    raise Exception(f"Build failed. Here is a link to the logs: {link}")
                return None
            else:
                logger.user_info(text=f"Building {docker_image_ref} docker image locally with {build_args}")
                image = docker.build(context_dir, file=file_path, load=True, tags=tags, **build_args)
                self._write_context_dir_checksum(self.build_dir / docker_image_ref, context_dir)
                return image  # type: ignore
        else:
            logger.user_info(
                text=f"Skipping build for context_dir={context_dir}, dockerfile={file_path} and image {tags[0]}"
            )
            return None

    def _pull_image(self, image_url: str) -> Union[Image, None]:
        if self.cloud_build or self.quick_mode:
            # TODO: verify that images exists remotely but dont pull them to local
            return None
        else:
            with logger.user_spinner("Pulling image locally"):
                image = docker.pull(image_url, quiet=True)
            return image  # type: ignore

    def find_image_model_by_name(self, image_name: str) -> DockerImageModel:
        """
        Finds a DockerImageModel with given image_name from self.image_models
        Args:
            image_name: name to find

        Returns:
            DockerImageModel
        """
        matched_image_models = list(filter(lambda i: i.name.strip() == image_name.strip(), self.image_models))
        if len(matched_image_models) == 0:
            raise ValueError(f"No docker image with name {image_name} found")
        elif len(matched_image_models) > 1:
            raise ValueError(f"Multiple docker images with name {image_name} found, please use unique names")
        else:
            return matched_image_models[0]

    def get_image(
        self,
        docker_image_ref: str,
    ) -> Tuple[DockerImageModel, Optional[Image], str]:
        """
        A wrapper around _get_image that checks if the docker image has been already build / pulled
        and cached in image_store.
        """
        if docker_image_ref in self.image_store:
            image = self.image_store.get(docker_image_ref)
            return image  # type: ignore
        else:
            image = self._get_image(
                docker_image_ref=docker_image_ref,
            )
            self.image_store.update({docker_image_ref: image})
            return image

    def _get_image(
        self,
        docker_image_ref: str,
    ) -> Tuple[DockerImageModel, Optional[Image], str]:
        """
        Given the docker_image_ref, this function prepares the image for you.
        Depending on the build_type, it either build the docker image or
        if you work with provided_image type, it will pull the image to verify the url.
        Args:
            docker_image_ref:

        Returns:

        """
        docker_image_model = self.find_image_model_by_name(docker_image_ref)

        build_dir = self.work_dir / Path("build") / "docker" / docker_image_model.name
        os.makedirs(build_dir, exist_ok=True)
        build_args = self.build_config.dict() if self.build_config else {}

        if isinstance(docker_image_model, NotebookReadyImageModel):
            tags = self.construct_image_tag(
                image_name=docker_image_model.name,
            )
            template_path = Path("notebook_template.Dockerfile")
            shutil.copy2(
                self.work_dir / docker_image_model.requirements_txt,
                build_dir / "requirements.txt",
            )
            file_path = self._jinja_render_dockerfile(docker_image_model, template_path, build_dir=build_dir)
            context_dir = build_dir
            image = self._build_image(
                context_dir, file_path=file_path, tags=tags, docker_image_ref=docker_image_ref, **build_args
            )
        elif isinstance(docker_image_model, LocalBuildImageModel):
            tags = self.construct_image_tag(
                image_name=docker_image_model.name,
            )
            file_path = self.work_dir / docker_image_model.dockerfile
            context_dir = self.work_dir / docker_image_model.context_dir
            image = self._build_image(
                context_dir, file_path=file_path, tags=tags, docker_image_ref=docker_image_ref, **build_args
            )
        elif isinstance(docker_image_model, ProvidedImageModel):
            image = self._pull_image(docker_image_model.image_url)
            tags = [docker_image_model.image_url]
        else:
            raise Exception("Invalid image model type.")

        return (
            docker_image_model,
            image,
            tags[0],
        )

    def _get_dirhash(self, directory: Path):
        dockerignore = directory / ".dockerignore"
        ignore = []

        if dockerignore.exists():
            with open(dockerignore, "r") as f:
                lines = f.readlines()
                ignore += [
                    ignore.rstrip() for ignore in lines if not ignore.startswith("#") and not ignore.strip() == ""
                ]

        return dirhash(directory, "sha256", ignore=set(ignore))

    def _get_cache_path(self, hash_cache_dir: Path):
        version = kebabcase(self.version)
        os.makedirs(hash_cache_dir, exist_ok=True)
        return hash_cache_dir / f"{kebabcase(self.docker_repository)}-{version}-cache.sha256"

    def _should_build_by_context_dir_checksum(self, hash_cache_dir: Path, context_dir: Path) -> bool:
        cache_file = self._get_cache_path(hash_cache_dir)
        sha256hash = self._get_dirhash(context_dir)
        if cache_file.exists():
            with open(cache_file, "r") as f:
                old_hash = f.read().replace("\n", "")
                return old_hash != sha256hash
        else:
            return True

    def _write_context_dir_checksum(self, hash_cache_dir: Path, context_dir: Path):
        cache_file = self._get_cache_path(hash_cache_dir)
        sha256hash = self._get_dirhash(context_dir)
        with open(cache_file, "w") as f:
            f.write(sha256hash)

    def _build_image_on_gcp_cloud_build(
        self, context_dir: Path, file_path: Path, tags: List[str], docker_image_ref: str
    ) -> Operation:
        """
        Build a docker container in GCP Cloud Build and push the images to registry.
        Folder context_dir is tarred, uploaded to GCS and then used to building.

        Args:
            context_dir: directory with all necessary files for docker image build
            file_path: path to Dockerfile
            tags:
        """

        dockerfile = os.path.relpath(file_path, context_dir)
        tar_filename = self.work_dir / "build" / "docker" / f"{docker_image_ref}.tar.gz"
        make_tarfile(context_dir, tar_filename)
        blob_name = os.path.relpath(tar_filename, self.work_dir).replace("\\", "/")
        blob = upload_file_to_gcs(filename=tar_filename, bucket_name=self.bucket, blob_name=blob_name)
        tags_args = " ".join([f"--destination={t}" for t in tags]).split()

        steps = BuildStep(
            name="gcr.io/kaniko-project/executor:latest",
            args=tags_args + ["--cache=true", "--dockerfile", dockerfile],
        )

        timeout = Duration()
        timeout.seconds = self.cloud_build_timeout

        # Set the pooling timeout to self.cloud_build_timeout seconds
        # since often large GPUs builds exceed the 900s limit
        DEFAULT_POLLING._timeout = self.cloud_build_timeout

        if self.cloud_build_workerpool:
            project_number = convert_project_id_to_project_number(self.project_id)
            options = BuildOptions(
                pool=BuildOptions.PoolOption(
                    name=f"projects/{project_number}/locations/{self.cloud_build_workerpool_location}"
                    f"/workerPools/{self.cloud_build_workerpool}"
                )
            )
            api_endpoint = f"{self.cloud_build_workerpool_location}-cloudbuild.googleapis.com"
        else:
            options = None
            api_endpoint = "cloudbuild.googleapis.com"

        build = Build(
            source=Source(storage_source=StorageSource(bucket=blob.bucket.name, object_=blob.name)),
            steps=[steps],
            # Issue with kaniko builder, images wont show in cloud build artifact column in UI
            # https://github.com/GoogleCloudPlatform/cloud-builders-community/issues/212
            # images=tags,
            timeout=timeout,
            options=options,
        )
        client = CloudBuildClient(
            credentials=get_credentials(), client_options=ClientOptions(api_endpoint=api_endpoint)
        )
        request = cloudbuild_v1.CreateBuildRequest(
            project_id=self.project_id,
            build=build,
        )
        return client.create_build(request=request)

    def push_image(self, image_or_tags: Union[Image, List[str]], quiet: bool = False) -> None:
        """
        Push a docker image to the registry (image must have tags)
        If you are in the cloud_build mode, nothing is pushed, images already live in cloud.
        Args:
            :param image_or_tags:
            :param quiet:
        """
        if not self.cloud_build:
            tags = image_or_tags.repo_tags if isinstance(image_or_tags, Image) else image_or_tags
            logger.user_info(text=f"Pushing docker image {tags}")
            docker.image.push(tags, quiet)

    def push_image_ref(self, image_ref: str, quiet: bool = False) -> None:
        """
        Push a docker image ref to the registry (image must have tags)
        Args:
            image_ref: image_ref to push
        """
        model, image, tag = self.get_image(image_ref)
        if (image or tag) and model.build_type != ImageBuildType.provided_image:
            self.push_image(image or [tag])

    @staticmethod
    def remove_image(image: Image, force=False, prune=True) -> None:
        """
        Remove docker image, useful if you dont want to clutter your machine.
        """
        docker.image.remove(image, force=force, prune=prune)

    def construct_image_tag(
        self,
        image_name: str,
    ):
        """
        Construct full image tag.
        Args:
            image_name:

        Returns:
            List of full image tag
        """
        versions = [self.version, "latest"]
        return [
            f"{self.docker_registry}/{self.docker_registry_suffix}{self.docker_project_id}/"
            f"{self.docker_repository}/{self.wanna_project_name}/{image_name}:{version}"
            for version in versions
        ]

    def build_container_and_get_image_url(self, docker_image_ref: str, push_mode: PushMode = PushMode.all) -> str:
        if push_mode == PushMode.quick:
            # only get image tag
            docker_image_model = self.find_image_model_by_name(docker_image_ref)
            tags = self.construct_image_tag(image_name=docker_image_model.name)
            image_url = tags[0]
        elif push_mode == PushMode.manifests:
            # only build image
            image_tag = self.get_image(docker_image_ref=docker_image_ref)
            image_url = image_tag[2]
        else:
            # build image and push
            image_tag = self.get_image(docker_image_ref=docker_image_ref)
            if len(image_tag) > 1 and image_tag[1]:
                self.push_image(image_tag[1])
            image_url = image_tag[2]
        return image_url

    @staticmethod
    def _jinja_render_dockerfile(
        image_model: DockerImageModel,
        template_path: Path,
        build_dir: Path,
    ) -> Path:
        """
        Based on image_model.type, we render dockerfile for this image type.
        Args:
            image_model: docker image model
            template_path: path to the dockerfile jinja template
            build_dir: build directory (where to save rendered dockerfile)

        Returns:
            path to the rendered dockerfile
        """
        if isinstance(image_model, NotebookReadyImageModel):
            rendered = render_template(
                source_path=template_path,
                base_image=image_model.base_image,
                requirements_txt=image_model.requirements_txt,
            )
        else:
            raise Exception("Invalid docker image type.")
        docker_file_path = build_dir / Path(f"{image_model.name}.Dockerfile")
        with open(docker_file_path, "w") as file:
            file.write(rendered)
        return docker_file_path



import itertools
from pathlib import Path
from threading import Thread
from typing import List, Tuple

import typer
from google.api_core import exceptions
from google.cloud.notebooks_v1.services.managed_notebook_service import ManagedNotebookServiceClient
from google.cloud.notebooks_v1.types import (
    ContainerImage,
    CreateRuntimeRequest,
    EncryptionConfig,
    LocalDisk,
    LocalDiskInitializeParams,
    Runtime,
    RuntimeAcceleratorConfig,
    RuntimeAccessConfig,
    RuntimeSoftwareConfig,
    VirtualMachine,
    VirtualMachineConfig,
)
from waiting import wait

from wanna.core.deployment.models import PushMode
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.notebook import ManagedNotebookModel
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.services.base import BaseService
from wanna.core.services.docker import DockerService
from wanna.core.services.tensorboard import TensorboardService
from wanna.core.utils import templates
from wanna.core.utils.config_enricher import email_fixer
from wanna.core.utils.gcp import upload_string_to_gcs

logger = get_logger(__name__)


class ManagedNotebookService(BaseService[ManagedNotebookModel]):
    def __init__(
        self,
        config: WannaConfigModel,
        workdir: Path,
        version: str = "dev",
    ):
        super().__init__(
            instance_type="managed-notebook",
        )
        self.version = version
        self.instances = config.managed_notebooks
        self.notebook_client = ManagedNotebookServiceClient()
        self.config = config
        self.bucket_name = config.gcp_profile.bucket
        self.tensorboard_service = TensorboardService(config=config)
        self.docker_service = (
            DockerService(
                docker_model=config.docker,
                gcp_profile=config.gcp_profile,
                version=version,
                work_dir=workdir,
                wanna_project_name=config.wanna_project.name,
            )
            if config.docker
            else None
        )

    def create(self, instance_name: str, **kwargs) -> None:
        """
        Create an instance with name "name" based on wanna-ml config.

        Args:
            instance_name: The name of the only instance from wanna-ml config that should be created.
                  Set to "all" to create everything from wanna-ml yaml configuration.
        """
        instances = self._filter_instances_by_name(instance_name)

        for instance in instances:
            t = Thread(target=self._create_one_instance, args=(instance,), kwargs={**kwargs})
            t.start()

    def delete(self, instance_name: str) -> None:
        """
        Delete an instance with name "name" based on wanna-ml config if exists on GCP.

        Args:
            instance_name: The name of the only instance from wanna-ml config that should be deleted.
                  Set to "all" to create all from configuration.
        """
        instances = self._filter_instances_by_name(instance_name)

        for instance in instances:
            t = Thread(target=self._delete_one_instance, args=(instance,))
            t.start()

    def _create_runtime_request(
        self, instance: ManagedNotebookModel, deploy: bool = True, push_mode: PushMode = PushMode.all
    ) -> CreateRuntimeRequest:
        """
        Transform the information about desired managed-notebook from the model based on yaml config
        to the form suitable for GCP API.

        Args:
            instance

        Returns:
            CreateRuntimeRequest
        """
        # Configuration of the managed notebook
        labels = {
            "wanna_name": instance.name,
            "wanna_resource": self.instance_type,
        }

        notebook_owner = instance.owner or self.config.gcp_profile.service_account

        if notebook_owner:
            labels["wanna_owner"] = email_fixer(notebook_owner)
            access_type = (
                RuntimeAccessConfig.RuntimeAccessType.SERVICE_ACCOUNT
                if notebook_owner.endswith("gserviceaccount.com")
                else RuntimeAccessConfig.RuntimeAccessType.SINGLE_USER
            )
        else:
            access_type = None

        if instance.labels:
            labels = {**instance.labels, **labels}

        # Disks
        disk_type = instance.data_disk.disk_type if instance.data_disk else None
        disk_size_gb = instance.data_disk.size_gb if instance.data_disk else None
        local_disk_params = LocalDiskInitializeParams(disk_size_gb=disk_size_gb, disk_type=disk_type)
        local_disk = LocalDisk(initialize_params=local_disk_params)
        encryption_config = (
            EncryptionConfig(kms_key=self.config.gcp_profile.kms_key) if self.config.gcp_profile.kms_key else None
        )

        # Accelerator
        if instance.gpu:
            runtimeAcceleratorConfig = RuntimeAcceleratorConfig(
                type=instance.gpu.accelerator_type, core_count=instance.gpu.count
            )
        else:
            runtimeAcceleratorConfig = None

        # Network and subnetwork
        full_network = self._get_resource_network(
            project_id=self.config.gcp_profile.project_id,
            push_mode=PushMode.all,
            resource_network=instance.network,
            fallback_project_network=self.config.gcp_profile.network,
            use_project_number=False,
        )
        subnet = instance.subnet if instance.subnet else self.config.gcp_profile.subnet
        full_subnet = self._get_resource_subnet(
            full_network,
            subnet,
            instance.region,
        )

        # Post startup script
        if deploy and instance.tensorboard_ref:
            script = self._prepare_startup_script(self.instances[0])
            blob = upload_string_to_gcs(
                script,
                instance.bucket or self.bucket_name,
                f"notebooks/{instance.name}/startup_script.sh",
            )
            post_startup_script = f"gs://{blob.bucket.name}/{blob.name}"
        else:
            post_startup_script = None

        # Runtime kernels
        kernels = []
        if instance.kernel_docker_image_refs:
            if not self.docker_service:
                raise Exception("Docker params in wanna-ml config not defined")
            for docker_image_ref in instance.kernel_docker_image_refs:
                image_url = self.docker_service.build_container_and_get_image_url(docker_image_ref, push_mode=push_mode)
                repository = image_url.partition(":")[0]
                tag = image_url.partition(":")[-1]
                kernels.append(ContainerImage(repository=repository, tag=tag))

        # VM
        virtualMachineConfig = VirtualMachineConfig(
            machine_type=instance.machine_type,
            container_images=kernels,
            data_disk=local_disk,
            encryption_config=encryption_config,
            labels=labels,
            accelerator_config=runtimeAcceleratorConfig,
            network=full_network,
            subnet=full_subnet,
            internal_ip_only=instance.internal_ip_only,
            tags=instance.tags,
            metadata=instance.metadata,
        )
        virtualMachine = VirtualMachine(virtual_machine_config=virtualMachineConfig)

        runtimeSoftwareConfig = RuntimeSoftwareConfig(
            {
                "post_startup_script": post_startup_script,
                "idle_shutdown": instance.idle_shutdown,
                "idle_shutdown_timeout": instance.idle_shutdown_timeout,
                "enable_health_monitoring": True,
            }
        )

        runtimeAccessConfig = RuntimeAccessConfig(access_type=access_type, runtime_owner=notebook_owner)
        runtime = Runtime(
            access_config=runtimeAccessConfig, software_config=runtimeSoftwareConfig, virtual_machine=virtualMachine
        )

        # Create runtime request
        return CreateRuntimeRequest(
            parent=f"projects/{instance.project_id}/locations/{instance.region}",
            runtime_id=instance.name,
            runtime=runtime,
        )

    def _delete_one_instance(self, notebook_instance: ManagedNotebookModel) -> None:
        """
        Delete one notebook instance. This assumes that it has been already verified that notebook exists.

        Args:
            notebook_instance: notebook to delete
        """

        exists = self._instance_exists(notebook_instance)
        if exists:
            logger.user_info(f"Deleting {self.instance_type} {notebook_instance.name} ...")
            deleted = self.notebook_client.delete_runtime(
                name=f"projects/{notebook_instance.project_id}/locations/"
                f"{notebook_instance.region}/runtimes/{notebook_instance.name}"
            )
            deleted.result()
            logger.user_success(f"Deleted {self.instance_type} {notebook_instance.name}")
        else:
            logger.user_error(
                f"Notebook with name {notebook_instance.name} was not found in region {notebook_instance.region}",
            )

    def _create_one_instance(self, instance: ManagedNotebookModel, **kwargs) -> None:
        """
        Create a notebook instance based on information in NotebookModel class.
        1. Check if the notebook already exists
        2. Parse the information from NotebookModel to GCP API friendly format = runtime_request
        3. Wait for the compute instance behind the notebook to start
        4. Wait for JupyterLab to start
        5. Get and print the link to JupyterLab

        Args:
            instance: notebook to be created

        """
        exists = self._instance_exists(instance)
        if exists:
            logger.user_info(f"Managed notebook {instance.name} already exists in location {instance.region}")
            should_recreate = typer.confirm("Are you sure you want to delete it and start a new?")
            if should_recreate:
                self._delete_one_instance(instance)
            else:
                return

        push_mode: PushMode = kwargs.get("push_mode")  # type: ignore
        request = self._create_runtime_request(instance=instance, deploy=True, push_mode=push_mode)
        logger.user_info(f"Creating instance for {instance.name} ...")
        nb_instance = self.notebook_client.create_runtime(request=request)
        instance_full_name = (
            nb_instance.result().name
        )  # .result() waits for compute engine behind the notebook to start
        logger.user_info(f"Starting JupyterLab for {instance.name} ...")
        wait(
            lambda: self._validate_jupyterlab_state(instance_full_name, Runtime.State.ACTIVE),
            timeout_seconds=450,
            sleep_seconds=20,
            waiting_for="Starting JupyterLab in your instance",
        )
        jupyterlab_link = self._get_jupyterlab_link(instance_full_name)
        logger.user_success(f"JupyterLab for {instance.name} started at {jupyterlab_link}")

    def _list_running_instances(self, project_id: str, location: str) -> List[str]:
        """
        List all notebooks with given project_id and location.

        Args:
            project_id: GCP project ID
            location: GCP location (zone)

        Returns:
            instance_names: List of the full names on notebook instances (this includes project_id, and zone)

        """
        instances = self.notebook_client.list_runtimes(parent=f"projects/{project_id}/locations/{location}")
        instance_names = [i.name for i in instances.runtimes]
        return instance_names

    def _instance_exists(self, instance: ManagedNotebookModel) -> bool:
        """
        Check if the instance with given instance_name exists in given GCP project project_id and location.
        Args:
            instance: notebook to verify if exists on GCP

        Returns:
            True if exists, False if not
        """
        full_instance_name = f"projects/{instance.project_id}/locations/{instance.region}/runtimes/{instance.name}"
        return full_instance_name in self._list_running_instances(
            instance.project_id, instance.region or self.config.gcp_profile.region
        )

    def _prepare_startup_script(self, nb_instance: ManagedNotebookModel) -> str:
        """
        Prepare the notebook startup script based on the information from notebook.
        This script run at the Compute Engine Instance creation time with a root user.

        Args:
            nb_instance

        Returns:
            startup_script
        """
        if nb_instance.tensorboard_ref:
            tensorboard_resource_name = self.tensorboard_service.get_or_create_tensorboard_instance_by_name(
                nb_instance.tensorboard_ref
            )
        else:
            tensorboard_resource_name = None

        startup_script = templates.render_template(
            Path("notebook_startup_script.sh.j2"),
            tensorboard_resource_name=tensorboard_resource_name,
        )
        return startup_script

    def _validate_jupyterlab_state(self, instance_id: str, state: int) -> bool:
        """
        Validate if the given notebook instance is in given state.

        Args:
            instance_id: Full notebook instance id
            state: Notebook state (ACTIVE, PENDING,...)

        Returns:
            True if desired state, False otherwise
        """
        try:
            instance_info = self.notebook_client.get_runtime(name=instance_id)
        except exceptions.NotFound:
            raise exceptions.NotFound(f"Managed Notebook {instance_id} was not found.") from None
        return instance_info.state == state

    def _get_jupyterlab_link(self, instance_id: str) -> str:
        """
        Get a link to jupyterlab proxy based on given notebook instance id.
        Args:
            instance_id: full notebook instance id

        Returns:
            proxy_uri: link to jupyterlab
        """
        instance_info = self.notebook_client.get_runtime({"name": instance_id})
        return f"https://{instance_info.access_config.proxy_uri}"

    def _return_diff(self) -> Tuple[List[ManagedNotebookModel], List[ManagedNotebookModel]]:
        """
        Figuring out the diff between GCP and wanna.yaml. Lists managed notebooks to be deleted and created.
        """
        location = self.config.gcp_profile.region
        parent = f"projects/{self.config.gcp_profile.project_id}/locations/{location}"
        active_runtimes = self.notebook_client.list_runtimes(parent=parent).runtimes
        wanna_notebook_names = [managednotebook.name for managednotebook in self.instances]
        active_notebook_names = [str(runtime.name) for runtime in active_runtimes]
        to_be_deleted = []
        to_be_created = []
        """
        Notebooks to be deleted
        """
        for runtime in active_runtimes:
            if runtime.virtual_machine.virtual_machine_config.labels["wanna_project"] == self.config.wanna_project.name:
                if runtime.name.split("/")[-1] not in wanna_notebook_names:
                    to_be_deleted.append(
                        ManagedNotebookModel.parse_obj(
                            {
                                "name": str(runtime.name).split("/")[-1],
                                "region": location,
                                "project_id": self.config.gcp_profile.project_id,
                                "owner": "wanna-to-be-deleted",
                            }
                        )
                    )
        """
        Notebooks to be created
        """
        for notebook in self.instances:
            if (
                f"projects/{notebook.project_id}/locations/{notebook.region}/runtimes/{notebook.name}"
                not in active_notebook_names
            ):
                to_be_created.append(notebook)

        return to_be_deleted, to_be_created

    def build(self) -> int:
        for instance in self.instances:
            self._create_runtime_request(instance=instance, deploy=False, push_mode=PushMode.manifests)
        logger.user_success("Managed notebooks validation OK!")
        return 0

    def push(self, instance_name: str):
        instances = self._filter_instances_by_name(instance_name)

        docker_image_refs = set(
            itertools.chain(
                *[instance.kernel_docker_image_refs for instance in instances if instance.kernel_docker_image_refs]
            )
        )
        if docker_image_refs:
            if self.docker_service:
                for docker_image_ref in docker_image_refs:
                    image_tag = self.docker_service.get_image(docker_image_ref=docker_image_ref)
                    if image_tag[1]:
                        self.docker_service.push_image(image_tag[1])
            else:
                raise Exception("Docker params in wanna-ml config not defined")



import os
from pathlib import Path
from string import Template
from typing import Union

import kfp.components as comp


def load_wanna_component(path: Union[Path, str]):
    with open(str(path), "r") as f:
        t = Template(f.read())
        component = t.safe_substitute(os.environ)
        return comp.load_component_from_text(component)



import importlib
import json
import os
from pathlib import Path
from typing import List, Optional, Tuple

from caseconverter import snakecase
from google.cloud import aiplatform
from kfp.v2.compiler import Compiler
from kfp.v2.compiler.main import compile_pyfile
from python_on_whales import Image

from wanna.core.deployment.artifacts_push import PushResult
from wanna.core.deployment.models import (
    ContainerArtifact,
    JsonArtifact,
    PathArtifact,
    PipelineResource,
    PushMode,
    PushTask,
)
from wanna.core.deployment.vertex_connector import VertexConnector
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.docker import DockerBuildResult, DockerImageModel, ImageBuildType
from wanna.core.models.pipeline import PipelineModel
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.services.base import BaseService
from wanna.core.services.docker import DockerService
from wanna.core.services.path_utils import PipelinePaths
from wanna.core.services.tensorboard import TensorboardService
from wanna.core.utils.loaders import load_yaml_path

logger = get_logger(__name__)


class PipelineService(BaseService[PipelineModel]):
    def __init__(
        self,
        config: WannaConfigModel,
        workdir: Path,
        version: str = "dev",
        push_mode: PushMode = PushMode.all,
        connector: VertexConnector[PipelineResource] = VertexConnector[PipelineResource](),
        kubeflow_pipeline_caching: Optional[bool] = None,
    ):
        super().__init__(
            instance_type="pipeline",
        )
        self.connector = connector
        self.instances = config.pipelines
        self.config = config
        self.workdir = workdir
        self.tensorboard_service = TensorboardService(config=config)
        self.version = version
        self.push_mode = push_mode
        self.docker_service = DockerService(
            docker_model=config.docker,  # type: ignore
            gcp_profile=config.gcp_profile,
            version=version,
            work_dir=workdir,
            wanna_project_name=self.config.wanna_project.name,
            quick_mode=push_mode.is_quick_mode(),
        )
        self.kubeflow_pipeline_caching = kubeflow_pipeline_caching
        self.notification_channels = {channel.name: channel for channel in self.config.notification_channels}

    def build(self, instance_name: str, pipeline_params_path: Optional[Path] = None) -> List[Path]:
        """
        Create an instance with name "name" based on wanna-ml config.
        Args:
            instance_name: The name of the only instance from wanna-ml config that should be created.
                  Set to "all" to create everything from wanna-ml yaml configuration.
            pipeline_params_path: Path with Kubeflow pipeline params,
                if present will override one from wanna.yaml
        """
        instances = self._filter_instances_by_name(instance_name)
        return [self._compile_one_instance(instance, pipeline_params_path) for instance in instances]

    def push(self, manifests: List[Path], local: bool = False) -> PushResult:
        return self.connector.push_artifacts(
            self.docker_service.push_image,
            self._prepare_push(manifests, self.version, local),
        )

    def _prepare_push(self, pipelines: List[Path], version: str, local: bool = False) -> List[PushTask]:
        push_tasks = []
        for local_manifest_path in pipelines:
            manifest = PipelineService.read_manifest(self.connector, str(local_manifest_path))
            pipeline_paths = PipelinePaths(self.workdir, manifest.pipeline_bucket, manifest.pipeline_name)
            json_artifacts, manifest_artifacts, container_artifacts = [], [], []

            logger.user_info(text=f"Packaging {manifest.pipeline_name} pipeline resources")

            # Push containers if we are running on Internal Teamcity build agent or on all push-mode
            if self.push_mode.can_push_containers():
                for ref in manifest.docker_refs:
                    if ref.build_type != ImageBuildType.provided_image:
                        container_artifacts.append(ContainerArtifact(name=ref.name, tags=ref.tags))

            # Push gcp resources if we are running on GCP build agent
            if self.push_mode.can_push_gcp_resources():
                # Prepare manifest paths
                local_kubeflow_json_spec_path = pipeline_paths.get_local_pipeline_json_spec_path(version)
                wanna_manifest_publish_path = pipeline_paths.get_gcs_wanna_manifest_path(version)
                kubeflow_json_spec_publish_path = pipeline_paths.get_gcs_pipeline_json_spec_path(version)

                if local:
                    # Override paths to local dir when in "local" mode, IE tests or local run
                    wanna_manifest_publish_path = pipeline_paths.get_local_wanna_manifest_path(version)
                    kubeflow_json_spec_publish_path = pipeline_paths.get_local_pipeline_json_spec_path(version)

                # Ensure to update manifest json_spec_path to have the actual gcs location
                manifest.json_spec_path = kubeflow_json_spec_publish_path

                json_artifacts.append(
                    JsonArtifact(
                        name="WANNA pipeline manifest",
                        json_body=manifest.dict(),
                        destination=wanna_manifest_publish_path,
                    )
                )

                manifest_artifacts.append(
                    PathArtifact(
                        name="Kubeflow V2 pipeline spec",
                        source=local_kubeflow_json_spec_path,
                        destination=manifest.json_spec_path,
                    )
                )

            push_tasks.append(
                PushTask(
                    manifest_artifacts=manifest_artifacts,
                    container_artifacts=container_artifacts,
                    json_artifacts=json_artifacts,
                )
            )

        return push_tasks

    @staticmethod
    def get_pipeline_bucket(bucket: Optional[str], fallback_bucket: str):
        if bucket:
            return bucket if bucket.startswith("gs://") else f"gs://{bucket}"
        else:
            return f"gs://{fallback_bucket}"

    def deploy(self, instance_name: str, env: str):
        instances = self._filter_instances_by_name(instance_name)
        for pipeline in instances:
            logger.user_info(f"Deploying {pipeline.name} version {self.version} to env {env}")
            pipeline_bucket = PipelineService.get_pipeline_bucket(pipeline.bucket, self.config.gcp_profile.bucket)
            pipeline_paths = PipelinePaths(self.workdir, pipeline_bucket, pipeline_name=pipeline.name)
            manifest = PipelineService.read_manifest(
                self.connector, pipeline_paths.get_gcs_wanna_manifest_path(self.version)
            )
            self.connector.deploy_pipeline(manifest, pipeline_paths, self.version, env)

    @staticmethod
    def run(
        pipelines: List[str],
        extra_params: Optional[Path] = None,
        sync: bool = True,
    ) -> None:
        connector = VertexConnector[PipelineResource]()
        for manifest_path in pipelines:
            manifest = PipelineService.read_manifest(connector, str(manifest_path))
            aiplatform.init(location=manifest.location, project=manifest.project)
            connector.run_pipeline(manifest, extra_params, sync)

    def _export_pipeline_params(
        self,
        pipeline_paths: PipelinePaths,
        pipeline_instance: PipelineModel,
        version: str,
        images: List[Tuple[DockerImageModel, Optional[Image], str]],
        tensorboard: Optional[str],
        network: Optional[str],
        pipeline_params_path: Optional[Path] = None,
    ):
        # Prepare env params to be exported
        pipeline_env_params = {
            "project_id": pipeline_instance.project_id,
            "pipeline_name": pipeline_instance.name,
            "version": version,
            "bucket": pipeline_instance.bucket,
            "region": pipeline_instance.region,
            "pipeline_root": pipeline_paths.get_gcs_pipeline_root(),
            "pipeline_labels": json.dumps(pipeline_instance.labels),
            "pipeline_service_account": (
                pipeline_instance.service_account
                if pipeline_instance.service_account
                else self.config.gcp_profile.service_account
            ),
            "pipeline_experiment": pipeline_instance.experiment,
        }

        if self.config.gcp_profile.kms_key:
            pipeline_env_params["encryption_spec_key_name"] = self.config.gcp_profile.kms_key

        if tensorboard:
            pipeline_env_params["tensorboard"] = tensorboard

        if network:
            pipeline_env_params["pipeline_network"] = network

        # Export Pipeline wanna ENV params to be available during compilation
        pipeline_name_prefix = snakecase(f"{pipeline_instance.name}").upper()
        for key, value in pipeline_env_params.items():
            env_name = snakecase(f"{pipeline_name_prefix}_{key.upper()}").upper()
            os.environ[env_name] = str(value)

        for (docker_image_model, _, tag) in images:
            env_name = snakecase(f"{docker_image_model.name}_DOCKER_URI").upper()
            os.environ[env_name] = tag

        # Collect pipeline compile params from wanna config
        if pipeline_params_path:
            pipeline_params_path = (self.workdir / pipeline_params_path).resolve()
            pipeline_compile_params = load_yaml_path(pipeline_params_path, self.workdir)
        else:
            if pipeline_instance.pipeline_params and isinstance(pipeline_instance.pipeline_params, Path):
                pipeline_params_path = (self.workdir / pipeline_instance.pipeline_params).resolve()
                pipeline_compile_params = load_yaml_path(pipeline_params_path, self.workdir)
            elif pipeline_instance.pipeline_params and isinstance(pipeline_instance.pipeline_params, dict):
                pipeline_compile_params = pipeline_instance.pipeline_params
            else:
                pipeline_compile_params = {}

        return pipeline_env_params, pipeline_compile_params

    def _compile_one_instance(self, pipeline: PipelineModel, pipeline_params_path: Optional[Path] = None) -> Path:

        image_tags = [
            self.docker_service.get_image(docker_image_ref=docker_image_ref)
            for docker_image_ref in pipeline.docker_image_ref
        ]

        logger.user_info(text=f"Compiling pipeline {pipeline.name}")

        # Prep build dir
        pipeline_bucket = PipelineService.get_pipeline_bucket(pipeline.bucket, self.config.gcp_profile.bucket)
        pipeline_paths = PipelinePaths(self.workdir, pipeline_bucket, pipeline.name)

        tensorboard = (
            self.tensorboard_service.get_or_create_tensorboard_instance_by_name(pipeline.tensorboard_ref)
            if pipeline.tensorboard_ref and self.push_mode.can_push_gcp_resources()
            else None
        )

        network = self._get_resource_network(
            project_id=pipeline.project_id,
            push_mode=self.push_mode,
            resource_network=pipeline.network,
            fallback_project_network=self.config.gcp_profile.network,
        )

        labels = {
            "wanna_resource_name": pipeline.name,
            "wanna_resource": self.instance_type,
        }
        if pipeline.sla_hours:
            labels["wanna_sla_hours"] = str(pipeline.sla_hours).replace(".", "_")
        if pipeline.labels:
            labels = {**pipeline.labels, **labels}
        pipeline.labels = labels
        # Collect kubeflow pipeline params for compilation
        pipeline_env_params, pipeline_params = self._export_pipeline_params(
            pipeline_paths, pipeline, self.version, image_tags, tensorboard, network, pipeline_params_path
        )

        # Compile kubeflow V2 Pipeline
        if pipeline.pipeline_function and not pipeline.pipeline_file:
            # This branch implies that the pipeline_function is
            # a python import path. ex: module1.module2.function
            mod_name, func_name = pipeline.pipeline_function.rsplit(".", 1)
            module = importlib.import_module(mod_name)
            logger.user_info(f"Using Compiler.compile with function {pipeline.pipeline_function}")
            func = getattr(module, func_name)
            Compiler().compile(
                pipeline_func=func,
                pipeline_parameters=pipeline_params,
                package_path=pipeline_paths.get_local_pipeline_json_spec_path(self.version),
                type_check=True,
            )
        elif pipeline.pipeline_file:
            # Get the file
            pyfile = str(self.workdir / pipeline.pipeline_file)
            logger.user_info(f"Using kfp compile_pyfile with {pyfile}")
            compile_pyfile(
                pyfile=pyfile,
                function_name=pipeline.pipeline_function,
                pipeline_parameters=pipeline_params,
                package_path=pipeline_paths.get_local_pipeline_json_spec_path(self.version),
                type_check=True,
                use_experimental=False,
            )
        else:
            raise ValueError("Can not compile kfp pipeline, " "pipeline_file or pipeline_function must be set.")

        docker_refs = [
            DockerBuildResult(
                name=model.name,
                tags=image.repo_tags if image and image.repo_tags else [tag],
                build_type=model.build_type,
            )
            for model, image, tag in image_tags
        ]

        channels = []
        for ref in pipeline.notification_channels_ref:
            channel = self.notification_channels.get(ref)
            if channel:
                channels.append(channel)
            else:
                raise ValueError(f"{ref} notification channel not specified")

        deployment_manifest = PipelineResource(
            name=f"pipeline {pipeline.name}",
            project=pipeline.project_id,
            location=pipeline.region if pipeline.region else self.config.gcp_profile.region,
            service_account=pipeline.service_account,
            network=network,
            pipeline_name=pipeline.name,
            pipeline_bucket=pipeline_bucket,
            pipeline_version=self.version,
            json_spec_path=pipeline_paths.get_local_pipeline_json_spec_path(self.version),
            parameter_values=pipeline_params,
            enable_caching=self.kubeflow_pipeline_caching or pipeline.enable_caching,
            labels=pipeline.labels,
            pipeline_root=pipeline_env_params.get("pipeline_root"),
            schedule=pipeline.schedule,
            docker_refs=docker_refs,
            compile_env_params=pipeline_env_params,
            notification_channels=channels,
            encryption_spec_key_name=self.config.gcp_profile.kms_key,
            experiment=pipeline.experiment,
        )

        manifest_path = pipeline_paths.get_local_wanna_manifest_path(self.version)
        self.connector.write(manifest_path, deployment_manifest.json())
        return Path(manifest_path).resolve()

    @staticmethod
    def read_manifest(connector: VertexConnector[PipelineResource], path: str) -> PipelineResource:
        return PipelineResource.parse_obj(connector.read(path))

    def _delete_one_instance(self, instance: PipelineModel) -> None:
        raise NotImplementedError

    def _create_one_instance(self, instance: PipelineModel, **kwargs) -> None:
        raise NotImplementedError

    def _instance_exists(self, instance: PipelineModel) -> bool:
        raise NotImplementedError



# Generated file do not change
import json
import os
from typing import Dict, Any

import pendulum
from google.cloud import aiplatform
from jinja2 import Environment

PROJECT_ID = os.getenv("PROJECT_ID")
REGION = os.getenv("REGION")
PIPELINE_ROOT = os.getenv("PIPELINE_ROOT")
PIPELINE_NETWORK = os.getenv("PIPELINE_NETWORK")
PIPELINE_SERVICE_ACCOUNT = os.getenv("PIPELINE_SERVICE_ACCOUNT")
PIPELINE_EXPERIMENT = os.getenv("PIPELINE_EXPERIMENT")
PIPELINE_LABELS = json.loads(os.environ["PIPELINE_LABELS"])  # if not define we won't run it
PIPELINE_JOB_ID = os.getenv("PIPELINE_JOB_ID")
ENCRYPTION_SPEC_KEY_NAME = os.getenv("ENCRYPTION_SPEC_KEY_NAME")

_jinja_env = Environment()
_jinja_env.globals.update(modules={"pendulum": pendulum})


def _update_time_template(params: Dict[str, Any]):

    for k, v in params.items():
        if isinstance(v, str):
            v = _jinja_env.from_string(v).render()
        params[k] = v

    return params


def process_request(request):
    """Processes the incoming HTTP request.

    Args:
      request (flask.Request): HTTP request object.

    Returns:
      The response text or any set of values that can be turned into a Response
      object using `make_response
      <http://flask.pocoo.org/docs/1.0/api/#flask.Flask.make_response>`.
    """

    # decode http request payload and translate into JSON object
    request_str = request.data.decode("utf-8")
    request_json = json.loads(request_str)

    pipeline_spec_uri = request_json["pipeline_spec_uri"]
    parameter_values = _update_time_template(request_json["parameter_values"])
    enable_caching = request_json.get("enable_caching", False)

    aiplatform.init(
        project=PROJECT_ID,
        location=REGION,
        experiment=PIPELINE_EXPERIMENT
    )

    job = aiplatform.PipelineJob(
        display_name="{{manifest.pipeline_name}}",
        job_id=PIPELINE_JOB_ID,
        template_path=pipeline_spec_uri,
        pipeline_root=PIPELINE_ROOT,
        enable_caching=enable_caching,
        parameter_values=parameter_values,
        labels=PIPELINE_LABELS,
        encryption_spec_key_name=ENCRYPTION_SPEC_KEY_NAME,
    )

    job.submit(service_account=PIPELINE_SERVICE_ACCOUNT,
               network=PIPELINE_NETWORK,
               experiment=PIPELINE_EXPERIMENT)

    return "Job submitted"



from datetime import datetime
from typing import Any, Dict

import pendulum
from jinja2 import Environment

_jinja_env = Environment()
_jinja_env.globals.update(modules={"pendulum": pendulum})


def get_timestamp():
    return datetime.now().strftime("%Y%m%d%H%M%S")


def update_time_template(params: Dict[str, Any]):

    for k, v in params.items():
        if isinstance(v, str):
            v = _jinja_env.from_string(v).render()
        params[k] = v

    return params



# mypy: allow-untyped-calls
import sys

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata

try:
    __version__ = importlib_metadata.version(__name__)
except importlib_metadata.PackageNotFoundError:
    __version__ = "snapshot"



import json
from typing import Any, Dict, Optional, cast

from google.cloud.exceptions import NotFound
from google.cloud.logging import Client as LoggingClient
from google.cloud.monitoring_v3 import (
    AlertPolicy,
    AlertPolicyServiceClient,
    NotificationChannel,
    NotificationChannelServiceClient,
)
from waiting import wait

from wanna.core.deployment.credentials import GCPCredentialsMixIn
from wanna.core.deployment.models import AlertPolicyResource, LogMetricResource, NotificationChannelResource
from wanna.core.loggers.wanna_logger import get_logger

logger = get_logger(__name__)


class MonitoringMixin(GCPCredentialsMixIn):
    @staticmethod
    def _get_notification_channel(
        client: NotificationChannelServiceClient, project: str, display_name: str
    ) -> Optional[NotificationChannel]:
        channels = list(client.list_notification_channels(name=f"projects/{project}"))
        channels = [channel for channel in channels if channel.display_name == display_name]
        if channels:
            return channels[0]
        return None

    def upsert_notification_channel(self, resource: NotificationChannelResource):
        client = NotificationChannelServiceClient(credentials=self.credentials)

        channel = MonitoringMixin._get_notification_channel(client, resource.project, resource.name)

        if not channel:
            notification_channel = NotificationChannel(
                type_=resource.type_,
                display_name=resource.name,
                description=resource.description,
                labels=resource.config,
                user_labels=resource.labels,
                verification_status=NotificationChannel.VerificationStatus.VERIFIED,
                enabled=True,
            )
            return client.create_notification_channel(
                name=f"projects/{resource.project}", notification_channel=notification_channel
            )
        else:
            return channel

    def upsert_alert_policy(self, resource: AlertPolicyResource):
        client = AlertPolicyServiceClient(credentials=self.credentials)

        alert_policy = {
            "display_name": resource.display_name,
            "user_labels": resource.labels,
            "conditions": [
                {
                    "display_name": "Error detected",
                    "condition_threshold": {
                        # https://issuetracker.google.com/issues/143436657?pli=1
                        # resource.type must be defined based on the resource type from log metric
                        "filter": f"""
                        metric.type="logging.googleapis.com/user/{resource.logging_metric_type}"
                        AND resource.type="{resource.resource_type}"
                        """,
                        "aggregations": [
                            {
                                "alignment_period": "600s",
                                "cross_series_reducer": "REDUCE_SUM",
                                "per_series_aligner": "ALIGN_DELTA",
                            }
                        ],
                        "comparison": "COMPARISON_GT",
                        "duration": "0s",
                        "trigger": {"count": 1},
                        "threshold_value": 0,
                    },
                }
            ],
            "alert_strategy": {
                "auto_close": "604800s",
            },
            "combiner": "OR",
            "enabled": True,
            "notification_channels": resource.notification_channels,
        }

        alert_policy = cast(AlertPolicy, AlertPolicy.from_json(json.dumps(alert_policy)))
        policies = client.list_alert_policies(name=f"projects/{resource.project}")
        policy = [policy for policy in policies if policy.display_name == resource.name]
        if policy:
            policy = policy[0]
            alert_policy.name = policy.name
            client.update_alert_policy(alert_policy=alert_policy)
        else:
            client.create_alert_policy(name=f"projects/{resource.project}", alert_policy=alert_policy)

    def upsert_log_metric(self, resource: LogMetricResource) -> Dict[str, Any]:
        client = LoggingClient(credentials=self.credentials)
        try:
            return client.metrics_api.metric_get(project=resource.project, metric_name=resource.name)
        except NotFound:
            client.metrics_api.metric_create(
                project=resource.project,
                metric_name=resource.name,
                filter_=resource.filter_,
                description=resource.description,
            )
            with logger.user_spinner(f"Creating log metric: {resource.name}"):
                wait(
                    lambda: client.metrics_api.metric_get(project=resource.project, metric_name=resource.name),
                    timeout_seconds=120,
                    sleep_seconds=5,
                    waiting_for="Log metric",
                )
            return client.metrics_api.metric_get(project=resource.project, metric_name=resource.name)



import logging
from typing import cast

from rich.console import Console
from rich.live import Live

if Console().encoding == "utf-8":
    in_progress_sign = ":hourglass_flowing_sand:"
    error_sign = ":x:"
    done_sign = ":white_check_mark:"
    info_sign = ":information_source:"
else:
    in_progress_sign = "(in progress)"
    error_sign = "(error)"
    done_sign = "(done)"
    info_sign = "(info)"


class Spinner(Live):
    def __init__(self, text: str, **kwargs):
        self.text = text
        super().__init__(text, **kwargs)

    def __enter__(self) -> Live:
        self.update(f"{in_progress_sign} {self.text}")
        self.start(refresh=self._renderable is not None)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Stops the spinner. For use in context managers."""
        if exception_value:
            self.update(f"{error_sign} {self.text}")
        else:
            self.update(f"{done_sign} {self.text}")
        self.stop()


class WannaLogger(logging.Logger):
    """
    This Logger supports all common logging library methods
    like .info, .warning or .debug.
    On top of that, we introduce new methods for visually
    more appealing printing to users.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Set the logging config here if needed
        logging.basicConfig()
        self.console = Console()

    def user_error(self, text) -> None:
        self.console.print(f"{error_sign} {text}", style="bold red")

    def user_info(self, text) -> None:
        self.console.print(f"{info_sign} {text}")

    def user_success(self, text) -> None:
        self.console.print(f"{done_sign} {text}")

    @staticmethod
    def user_spinner(text, **kwargs) -> Spinner:
        return Spinner(text=text, **kwargs)


def get_logger(name: str) -> WannaLogger:
    logging.setLoggerClass(WannaLogger)
    logger = cast(WannaLogger, logging.getLogger(name))
    return logger



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import typer
from google.cloud import aiplatform
from google.cloud.aiplatform import CustomTrainingJob
from google.cloud.aiplatform.gapic import WorkerPoolSpec
from google.cloud.aiplatform_v1.types import ContainerSpec, DiskSpec, MachineSpec, PythonPackageSpec
from google.cloud.aiplatform_v1.types.pipeline_state import PipelineState
from google.protobuf.json_format import MessageToDict

from wanna.core.deployment.models import ContainerArtifact, JobResource, PathArtifact, PushMode, PushResult, PushTask
from wanna.core.deployment.vertex_connector import VertexConnector
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.docker import ImageBuildType
from wanna.core.models.training_custom_job import (
    CustomJobModel,
    HyperparameterTuning,
    JobModelTypeAlias,
    TrainingCustomJobModel,
    WorkerPoolModel,
)
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.services.base import BaseService, T
from wanna.core.services.docker import DockerService
from wanna.core.services.path_utils import JobPaths
from wanna.core.services.tensorboard import TensorboardService
from wanna.core.utils.json import remove_nones
from wanna.core.utils.loaders import load_yaml_path

logger = get_logger(__name__)


class JobService(BaseService[JobModelTypeAlias]):
    def __init__(
        self,
        config: WannaConfigModel,
        workdir: Path,
        version: str = "dev",
        push_mode: PushMode = PushMode.all,
        connector: VertexConnector[JobResource[JobModelTypeAlias]] = VertexConnector[JobResource[JobModelTypeAlias]](),
    ):
        """
        Service to build, push, deploy and run Vertex AI custom jobs
        Args:
            config (WannaConfigModel): Loaded wanna.yaml
            workdir (Path): Where wanna will conduct it's work
            version (str): Which version of the jobs are working with
        """
        super().__init__(
            instance_type="job",
        )

        self.connector = connector
        self.instances = config.jobs
        self.wanna_project = config.wanna_project
        self.bucket_name = config.gcp_profile.bucket
        self.config = config
        self.tensorboard_service = TensorboardService(config=config)
        self.push_mode = push_mode
        self.workdir = workdir
        self.docker_service = DockerService(
            docker_model=config.docker,  # type: ignore
            gcp_profile=config.gcp_profile,
            version=version,
            work_dir=workdir,
            wanna_project_name=self.wanna_project.name,
            quick_mode=self.push_mode.is_quick_mode(),
        )
        self.build_dir = workdir / "build"
        self.version = version

    def build(self, instance_name: str) -> List[Path]:
        """
        Based on wanna config and setup it creates a JobManifest that
        can later be pushed, deployed or run
        Args:
            instance_name: the give job(s) that will be built
                "all" means it will build all jobs

        Returns:
            Job Manifests and associated local paths where those were built
        """
        instances = self._filter_instances_by_name(instance_name)
        return [self._build(instance) for instance in instances]

    def push(self, manifests: List[Path], local: bool = False) -> PushResult:
        return self.connector.push_artifacts(
            self.docker_service.push_image, self._prepare_push(manifests, self.version, local)
        )

    def _prepare_push(self, manifests: List[Path], version: str, local: bool = False) -> List[PushTask]:
        """
        Completes the build process buy pushing docker images and pushing manifest files to
        GCS for future execution
        Args:
            manifests: Job Manifests and associated local paths where those were built
            local: allows to publish the manifests locally for inspection and tests

        Returns:
            paths to gs:// paths where the manifests were pushed
        """

        push_tasks = []
        for manifest in manifests:
            manifest_path = str(manifest.resolve())
            loaded_manifest = JobService.read_manifest(self.connector, manifest_path)
            job_paths = JobPaths(self.workdir, f"gs://{self.bucket_name}", loaded_manifest.job_config.name)
            manifest_artifacts, container_artifacts = [], []

            logger.user_info(f"Packaging {loaded_manifest.job_config.name} job resources")

            if self.push_mode.can_push_containers():
                for docker_image_ref in loaded_manifest.image_refs:
                    model, image, tag = self.docker_service.get_image(docker_image_ref)
                    if model.build_type != ImageBuildType.provided_image:
                        tags = image.repo_tags if image and image.repo_tags else [tag]
                        container_artifacts.append(ContainerArtifact(name=model.name, tags=tags))

            if self.push_mode.can_push_gcp_resources():
                gcs_manifest_path = job_paths.get_gcs_job_wanna_manifest_path(version)

                if not local:
                    manifest_artifacts.append(
                        PathArtifact(
                            name=f"{loaded_manifest.job_config.name} job manifest",
                            source=manifest_path,
                            destination=gcs_manifest_path,
                        )
                    )

            push_tasks.append(
                PushTask(
                    container_artifacts=container_artifacts,
                    manifest_artifacts=manifest_artifacts,
                    json_artifacts=[],
                )
            )

        return push_tasks

    @staticmethod
    def run(
        manifests: List[str],
        sync: bool = True,
        hp_params: Optional[Path] = None,
        command_override: Optional[List[str]] = None,
        args_override: Optional[List[Union[str, float, int]]] = None,
    ) -> None:
        """
        Run a Vertex AI Custom Job(s) with a given JobManifest
        Args:
            manifests (List[str]): WANNA JobManifests to be executed
            sync (bool): Allows to run the job in async vs sync mode
            hp_params: path with yaml hp_params for the job
            command_override: override default command from wanna.yaml.
                allows for quickly run a job with multiple commands permutations
            args_override:
                allows for quickly run a job with different args

        """

        for manifest_path in manifests:
            connector = VertexConnector[JobResource[JobModelTypeAlias]]()
            manifest = JobService.read_manifest(connector, manifest_path)

            if isinstance(manifest.job_config, TrainingCustomJobModel):
                if args_override:
                    manifest.job_config.worker.args = args_override

                if manifest.job_config.worker.container and command_override:
                    manifest.job_config.worker.container.command = command_override
                    manifest.job_payload["command"] = command_override

                if manifest.job_config.worker.python_package and command_override:
                    manifest.job_config.worker.python_package.module_name = " ".join(command_override)
                    manifest.job_payload["python_module_name"] = " ".join(command_override)

            elif isinstance(manifest.job_config, CustomJobModel):
                logger.user_info(
                    "command and args override is not supported in CustomJobModel jobs with multiple workers"
                )

            aiplatform.init(location=manifest.job_config.region, project=manifest.job_config.project_id)

            if isinstance(manifest.job_config, CustomJobModel):
                if hp_params:
                    override_hp_params = load_yaml_path(hp_params, Path("."))
                    manifest_hp_params = manifest.job_config.hp_tuning.dict() if manifest.job_config.hp_tuning else {}
                    overriden_hp_tuning = HyperparameterTuning.parse_obj({**manifest_hp_params, **override_hp_params})
                    manifest.job_config.hp_tuning = overriden_hp_tuning
                connector.run_custom_job(manifest, sync)
            else:
                connector.run_training_job(manifest, sync)

    def _build(self, instance: Union[CustomJobModel, TrainingCustomJobModel]) -> Path:
        """
        Creates a JobManifest that can later be pushed, deployed or run

        Args:
            instance: custom job model to create
        Returns:
            Job Resource for execution
        """
        job_paths = JobPaths(self.workdir, instance.bucket or f"gs://{self.config.gcp_profile.bucket}", instance.name)
        manifest_path = Path(job_paths.get_local_job_wanna_manifest_path(self.version))
        resource: Union[JobResource[CustomJobModel], JobResource[TrainingCustomJobModel]] = (
            self._create_training_job_resource(instance)
            if isinstance(instance, TrainingCustomJobModel)
            else self._create_custom_job_resource(instance)
        )

        return self.write_manifest(manifest_path, resource)

    def _create_custom_job_resource(
        self,
        job_model: CustomJobModel,
    ) -> JobResource[CustomJobModel]:
        image_refs, worker_pool_specs = set(
            zip(*[self._create_worker_pool_spec(worker) for worker in job_model.workers])
        )
        labels = {
            "wanna_name": job_model.name,
            "wanna_resource": self.instance_type,
        }
        if job_model.labels:
            labels = {**job_model.labels, **labels}

        network = self._get_resource_network(
            project_id=self.config.gcp_profile.project_id,
            push_mode=self.push_mode,
            resource_network=job_model.network,
            fallback_project_network=self.config.gcp_profile.network,
        )

        encryption_spec_key_name = (
            job_model.encryption_spec if job_model.encryption_spec else self.config.gcp_profile.kms_key
        )
        env_vars = self.config.gcp_profile.env_vars if self.config.gcp_profile.env_vars else dict()
        if job_model.env_vars:
            env_vars = {**env_vars, **job_model.env_vars}
        return JobResource[CustomJobModel](
            name=job_model.name,
            project=job_model.project_id,
            location=job_model.region,
            job_config=job_model,
            job_payload={
                "display_name": job_model.name,
                "worker_pool_specs": [
                    remove_nones(MessageToDict(s._pb, preserving_proto_field_name=True))
                    for s in list(worker_pool_specs)
                ],
                "labels": labels,
                "staging_bucket": job_model.bucket,
            },
            image_refs=list(image_refs),
            # during `run` calls, this means changing TensorboardService init
            tensorboard=self.tensorboard_service.get_or_create_tensorboard_instance_by_name(job_model.tensorboard_ref)
            if job_model.tensorboard_ref
            else None,
            network=network,
            encryption_spec=encryption_spec_key_name,
            environment_variables=env_vars,
        )

    def _create_training_job_resource(
        self,
        job_model: TrainingCustomJobModel,
    ) -> JobResource[TrainingCustomJobModel]:
        """
        Creates a Training Job Manifest that can later be pushed, deployed or run

        Args:
            job_model: Parsed TrainingCustomJobModel from wanna.yaml
        Returns:
            Custom Python on Custom Container training job Manifests
        """
        labels = {
            "wanna_name": job_model.name,
            "wanna_resource": self.instance_type,
        }
        if job_model.labels:
            labels = {**job_model.labels, **labels}
        job_payload: Dict[str, Any] = {}
        if job_model.worker.python_package:
            image_ref = job_model.worker.python_package.docker_image_ref
            _, _, tag = self.docker_service.get_image(docker_image_ref=job_model.worker.python_package.docker_image_ref)
            job_payload = {
                "display_name": job_model.name,
                "python_package_gcs_uri": job_model.worker.python_package.package_gcs_uri,
                "python_module_name": job_model.worker.python_package.module_name,
                "container_uri": tag,
                "labels": labels,
                "staging_bucket": job_model.bucket,
            }
        elif job_model.worker.container:
            image_ref = job_model.worker.container.docker_image_ref
            _, _, tag = self.docker_service.get_image(docker_image_ref=job_model.worker.container.docker_image_ref)
            job_payload = {
                "display_name": job_model.name,
                "container_uri": tag,
                "command": job_model.worker.container.command,
                "labels": labels,
                "staging_bucket": job_model.bucket,
            }
        else:
            raise ValueError(f"Job {job_model.name} worker must have `container` or `python_package` defined")

        network = self._get_resource_network(
            project_id=self.config.gcp_profile.project_id,
            push_mode=self.push_mode,
            resource_network=job_model.network,
            fallback_project_network=self.config.gcp_profile.network,
        )
        encryption_spec_key_name = (
            job_model.encryption_spec if job_model.encryption_spec else self.config.gcp_profile.kms_key
        )
        env_vars = self.config.gcp_profile.env_vars if self.config.gcp_profile.env_vars else dict()
        if job_model.env_vars:
            env_vars = {**env_vars, **job_model.env_vars}
        return JobResource[TrainingCustomJobModel](
            name=job_model.name,
            project=job_model.project_id,
            location=job_model.region,
            job_config=job_model,
            job_payload=job_payload,
            image_refs=[image_ref],
            tensorboard=self.tensorboard_service.get_or_create_tensorboard_instance_by_name(job_model.tensorboard_ref)
            if job_model.tensorboard_ref
            else None,
            network=network,
            encryption_spec=encryption_spec_key_name,
            environment_variables=env_vars,
        )

    def _create_worker_pool_spec(self, worker_pool_model: WorkerPoolModel) -> Tuple[str, WorkerPoolSpec]:
        """
        Converts the friendlier WANNA WorkerPoolModel to aiplatform sdk equivalent
        Args:
            worker_pool_model: Wanna user specified pool details

        Returns:
            The wanna container image_ref to be pushed
            and the aiplatform sdk worker pool spec
        """

        if worker_pool_model.container:
            image_ref = worker_pool_model.container.docker_image_ref
        elif worker_pool_model.python_package:
            image_ref = worker_pool_model.python_package.docker_image_ref
        else:
            raise ValueError(
                "Worker pool does not have container nor python_package. " "This means validation has a bug."
            )

        return image_ref, WorkerPoolSpec(
            container_spec=ContainerSpec(
                image_uri=self.docker_service.get_image(image_ref)[2],
                command=worker_pool_model.container.command,
                args=worker_pool_model.args,
                env=worker_pool_model.args,
            )
            if worker_pool_model.container
            else None,
            python_package_spec=PythonPackageSpec(
                executor_image_uri=self.docker_service.get_image(image_ref)[2],
                package_uris=[worker_pool_model.python_package.package_gcs_uri],
                python_module=worker_pool_model.python_package.module_name,
            )
            if worker_pool_model.python_package
            else None,
            machine_spec=MachineSpec(
                machine_type=worker_pool_model.machine_type,
                accelerator_type=worker_pool_model.gpu.accelerator_type if worker_pool_model.gpu else None,
                accelerator_count=worker_pool_model.gpu.count if worker_pool_model.gpu else None,
            ),
            disk_spec=DiskSpec(
                boot_disk_type=worker_pool_model.boot_disk.disk_type,
                boot_disk_size_gb=worker_pool_model.boot_disk.size_gb,
            )
            if worker_pool_model.boot_disk
            else None,
            replica_count=worker_pool_model.replica_count,
        )

    @staticmethod
    def _create_list_jobs_filter_expr(states: List[PipelineState], job_name: Optional[str] = None) -> str:
        """
        Creates a filter expression that can be used when listing current jobs on GCP.
        Args:
            states: list of desired states
            job_name: desire job name

        Returns:
            filter expression
        """
        filter_expr = "(" + " OR ".join([f'state="{state.name}"' for state in states]) + ")"
        if job_name:
            filter_expr = filter_expr + f' AND display_name="{job_name}"'
        return filter_expr

    def _list_jobs(self, states: List[PipelineState], job_name: Optional[str] = None) -> List[CustomTrainingJob]:
        """
        List all custom jobs with given project_id, region with given states.

        Args:
            states: list of custom job states, eg [JobState.JOB_STATE_RUNNING, JobState.JOB_STATE_PENDING]
            job_name:

        Returns:
            list of jobs
        """
        filter_expr = self._create_list_jobs_filter_expr(states=states, job_name=job_name)
        jobs = aiplatform.CustomTrainingJob.list(filter=filter_expr)
        return jobs  # type: ignore

    def _stop_one_instance(self, instance: Union[CustomJobModel, TrainingCustomJobModel]) -> None:
        """
        Pause one all jobs that have the same region and name as "instance".
        First we list all jobs with state running and pending and then
        user is prompted to choose which to kill.

        Args:
            instance: custom job model
        """
        active_jobs = self._list_jobs(
            states=[PipelineState.PIPELINE_STATE_RUNNING, PipelineState.PIPELINE_STATE_PENDING],  # type: ignore
            job_name=instance.name,
        )
        if active_jobs:
            for job in active_jobs:
                should_cancel = typer.prompt(
                    f"Do you want to cancel job {job.display_name} (started at {job.create_time})?"
                )
                if should_cancel:
                    with logger.user_spinner(f"Canceling job {job.display_name}"):
                        job.cancel()
        else:
            logger.user_info(f"No running or pending job with name {instance.name}")

    @staticmethod
    def read_manifest(
        connector: VertexConnector[JobResource[JobModelTypeAlias]],
        manifest_path: str,
    ) -> Union[JobResource[CustomJobModel], JobResource[TrainingCustomJobModel]]:
        """
        Reads a job manifest file

        Args:
            manifest_path (Path): Manifest path to be loaded
            connector VertexConnector[JobResource[JOB]]: connector to read manifest

        Returns:
            JobManifest: Parsed and loaded JobManifest

        """

        json_dict = connector.read(manifest_path)
        try:
            return JobResource[CustomJobModel].parse_obj(json_dict)
        except:
            return JobResource[TrainingCustomJobModel].parse_obj(json_dict)

    def write_manifest(
        self,
        local_manifest_path: Path,
        resource: Union[JobResource[TrainingCustomJobModel], JobResource[CustomJobModel]],
    ) -> Path:
        """
        Writes a JobManifest to a local path

        Args:
            local_manifest_path (Path): wanna local path save manifest
            resource (JobResource): the job resource manifest that should be saved locally

        Returns:
            Path: Path where resource manifest was saved to
        """
        encryption_spec_key_name = (
            resource.encryption_spec if resource.encryption_spec else self.config.gcp_profile.kms_key
        )
        env_vars = self.config.gcp_profile.env_vars if self.config.gcp_profile.env_vars else dict()
        if resource.environment_variables:
            env_vars = {**env_vars, **resource.environment_variables}
        json_dict = {
            "name": resource.name,
            "project": resource.project,
            "location": resource.location,
            "job_config": resource.job_config.dict(),
            "image_refs": resource.image_refs,
            "job_payload": resource.job_payload,
            "tensorboard": resource.tensorboard,
            "network": resource.network,
            "encryption_spec": encryption_spec_key_name,
            "environment_variables": env_vars,
        }
        json_dump = json.dumps(
            remove_nones(json_dict),
            allow_nan=False,
            default=lambda o: dict((key, value) for key, value in o.__dict__.items() if value),
        )
        self.connector.write(local_manifest_path, json_dump)

        return local_manifest_path

    def _delete_one_instance(self, instance: T) -> None:
        raise NotImplementedError

    def _create_one_instance(self, instance: T, **kwargs) -> None:
        raise NotImplementedError

    def _return_diff(self) -> Tuple[List[T], List[T]]:
        raise NotImplementedError
